# -*- coding: utf-8 -*-

# Copyright (c) 2016 HashData Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from __future__ import unicode_literals

from datetime import timedelta
import struct

from pytidbrep.expcetion import InvalidRowData
from pytidbrep.pb_binlog.binlog_pb2 import DDL
from pytidbrep.pb_binlog.binlog_pb2 import DML
from pytidbrep.pb_binlog.binlog_pb2 import Delete
from pytidbrep.pb_binlog.binlog_pb2 import Insert
from pytidbrep.pb_binlog.binlog_pb2 import Update
from pytidbrep.pb_binlog.binlog_pb2 import column

SIGNMASK = 0x8000000000000000

DIGITS_PER_WORD = 9  # A word holds 9 digits.
WORD_SIZE = 4  # A word is 4 bytes int32.
DIG2BYTES = [0, 1, 1, 2, 2, 3, 3, 4, 4, 4]

TYPE_BIT = 16
TYPE_TINY = 1
TYPE_SHORT = 2
TYPE_INT24 = 9
TYPE_LONG = 3
TYPE_LONGLONG = 8
TYPE_FLOAT = 4
TYPE_DOUBLE = 5
TYPE_DECIMAL = 0
TYPE_NEWDECIMAL = 0xf6

TYPE_DATE = 10
TYPE_NEWDATE = 14
TYPE_DATETIME = 12
TYPE_TIMESTAMP = 7
TYPE_YEAR = 13
TYPE_DURATION = 11

TYPE_VARCHAR = 15
TYPE_VARSTRING = 0xfd
TYPE_STRING = 0xfe
TYPE_TINYBLOB = 0xf9
TYPE_BLOB = 0xfc
TYPE_MEDIUMBLOB = 0xfa
TYPE_LONGBLOB = 0xfb
TYPE_ENUM = 0xf7
TYPE_SET = 0xf8

TYPE_NULL = 6
TYPE_GEOMETRY = 0xff

MYSQL_TYPE_BIT_STR = "bit"
MYSQL_TYPE_TINYINT = "tinyint"
MYSQL_TYPE_SMALLINT = "smallint"
MYSQL_TYPE_MEDIUMINT = "mediumint"
MYSQL_TYPE_INT = "int"
MYSQL_TYPE_BIGINT = "bigint"
MYSQL_TYPE_DECIMAL = "decimal"
MYSQL_TYPE_FLOAT = "float"
MYSQL_TYPE_DOUBLE = "double"

MYSQL_TYPE_DATE = "date"
MYSQL_TYPE_DATETIME = "datetime"
MYSQL_TYPE_TIMESTAMP = "timestamp"
MYSQL_TYPE_TIME = "time"
MYSQL_TYPE_YEAR = "year"

MYSQL_TYPE_CHAR = "char"
MYSQL_TYPE_VARCHAR = "varchar"
MYSQL_TYPE_BINARY = "binary"
MYSQL_TYPE_VARBINARY = "varbinary"
MYSQL_TYPE_TINYBLOB = "tinyblob"
MYSQL_TYPE_TINYTEXT = "tinytext"
MYSQL_TYPE_BLOB = "blob"
MYSQL_TYPE_TEXT = "text"
MYSQL_TYPE_MEDIUMBLOB = "mediumblob"
MYSQL_TYPE_MEDIUMTEXT = "mediumtext"
MYSQL_TYPE_LONGBLOB = "longblob"
MYSQL_TYPE_LONGTEXT = "longtext"
MYSQL_TYPE_ENUM = "enum"

MYSQL_TYPE_GEOMETRY = "geometry"
MYSQL_TYPE_NULL = "null"
MYSQL_TYPE_VARSTRING = "var_string"
MYSQL_TYPE_UNSPECIFIED = "unspecified"


def get_unicode(s):
    if isinstance(s, unicode):
        return s
    else:
        return s.decode('utf8')


def format_column(t, v):
    _, mysql_type = t

    if v is None:
        return "NULL"
    elif mysql_type in (MYSQL_TYPE_TINYTEXT, MYSQL_TYPE_MEDIUMTEXT,
                        MYSQL_TYPE_LONGTEXT, MYSQL_TYPE_TEXT, MYSQL_TYPE_CHAR,
                        MYSQL_TYPE_VARCHAR):
        return '"%s"' % get_unicode(v)
    elif mysql_type in (MYSQL_TYPE_BINARY, MYSQL_TYPE_VARBINARY,
                        MYSQL_TYPE_TINYBLOB, MYSQL_TYPE_MEDIUMBLOB,
                        MYSQL_TYPE_LONGBLOB, MYSQL_TYPE_BLOB):
        return '"%s"' % v.encode('hex')
    elif mysql_type in (MYSQL_TYPE_DATE, MYSQL_TYPE_DATETIME,
                        MYSQL_TYPE_TIMESTAMP, MYSQL_TYPE_TIME,
                        MYSQL_TYPE_YEAR):
        return '"%s"' % str(v)
    else:
        return str(v)


def int2byte(i):
    return struct.pack(b"!B", i)


def read_be_uint64(buf):
    return struct.unpack(b">Q", buf)[0]


def read_int8(buf):
    return struct.unpack(b">b", buf)[0]


def read_uint8(buf):
    return struct.unpack(b">B", buf)[0]


def read_uvarint(buf):
    '''
     read_uvarint decodes a uint64 from buf and returns that value and the
     number of bytes read (> 0). If an error occurred, the value is 0
     and the number of bytes n is <= 0 meaning:

        n == 0: buf too small
        n  < 0: value larger than 64 bits (overflow)
                  and -n is the number of bytes read
    '''

    x = 0
    s = 0

    for i in range(len(buf)):
        b = read_uint8(buf[i])

        if b < 0x80:
            if i > 9 or i == 9 and b > 1:
                return 0, -(i + 1)  # overflow

            return x | (b << s), i + 1

        x |= (b & 0x7f) << s
        s += 7

    return 0, 0


def read_varint(buf):
    '''
     read_varint decodes an int64 from buf and returns that value and the
     number of bytes read (> 0). If an error occurred, the value is 0
     and the number of bytes n is <= 0 with the following meaning:

        n == 0: buf too small
        n  < 0: value larger than 64 bits (overflow)
                  and -n is the number of bytes read
    '''

    ux, n = read_uvarint(buf)  # ok to continue in presence of error

    x = ux >> 1

    if ux & 1 != 0:
        x = ~x

    return x, n


def read_int32_word(buf, size):
    if read_int8(buf[0]) & 0x80 > 0:
        pad = 0xFF
    else:
        pad = 0

    tmp = bytearray(WORD_SIZE)

    for i in range(WORD_SIZE - size):
        tmp[i] = pad

    offset = WORD_SIZE - size
    for i in range(offset, WORD_SIZE):
        tmp[i] = buf[i - offset]

    x = struct.unpack(b">i", str(tmp))[0]

    return x


def decimal_bin_size(precision, frac):
    digits_int = precision - frac
    words_int = digits_int / DIGITS_PER_WORD
    words_frac = frac / DIGITS_PER_WORD
    xint = digits_int - words_int * DIGITS_PER_WORD
    xfrac = frac - words_frac * DIGITS_PER_WORD
    return words_int * WORD_SIZE + DIG2BYTES[xint] + \
        words_frac * WORD_SIZE + DIG2BYTES[xfrac]


class BinLogEvent(object):
    def __init__(self, binlog):
        self.tp = binlog.tp
        self.commit_ts = binlog.commit_ts

    @classmethod
    def type_name(cls, tp):
        if tp == DML:
            return "DML"
        elif tp == DDL:
            return "DDL"
        else:
            return "Unknown"

    def __str__(self):
        return unicode(self).encode('utf-8')

    def __unicode__(self):
        return "%s: %s" % (self.commit_ts, self.type_name(self.tp))


class DDLEvent(BinLogEvent):
    def __init__(self, binlog):
        super(DDLEvent, self).__init__(binlog)
        self._statement = get_unicode(binlog.ddl_query)

    @property
    def statement(self):
        '''Return DDL statement

        @return: A unicode string of DDL statement
        '''
        return self._statement

    def __str__(self):
        return unicode(self).encode('utf-8')

    def __unicode__(self):
        return "%s: %s" % (super(DDLEvent, self).__unicode__(), self.statement)


class XidEvent(BinLogEvent):
    """A COMMIT event
    """

    def __init__(self, ts):
        self.commit_ts = ts

    def __str__(self):
        return unicode(self).encode('utf-8')

    def __unicode__(self):
        return "Transaction at: %s" % (self.commit_ts)


class RowsEvent(BinLogEvent):
    NIL_FLAG = 0
    BYTES_FLAG = 1
    COMPACTBYTES_FLAG = 2
    INT_FLAG = 3
    UINT_FLAG = 4
    FLOAT_FLAG = 5
    DECIMAL_FLAG = 6
    DURATION_FLAG = 7
    VARINT_FLAG = 8
    UVARINT_FLAG = 9
    MAX_FLAG = 250

    def __init__(self, binlog, event):
        super(RowsEvent, self).__init__(binlog)
        self._schema = get_unicode(event.schema_name)
        self._table = get_unicode(event.table_name)
        self._dml_tp = event.tp

    @property
    def schema(self):
        '''Return table's schema name of the event

        @return: A unicode string of schema
        '''
        return self._schema

    @property
    def table(self):
        '''Return table's name of the event

        @return: A unicode string of table name
        '''
        return self._table

    @classmethod
    def dml_type_name(cls, tp):
        if tp == Insert:
            return "INSERT"
        elif tp == Update:
            return "UPDATE"
        elif tp == Delete:
            return "DELETE"
        else:
            return "UNKNOWN"

    def __str__(self):
        return unicode(self).encode('utf-8')

    def __unicode__(self):
        parent = super(RowsEvent, self).__unicode__()
        return "%s: %s %s.%s" % (parent, self.dml_type_name(self._dml_tp),
                                 self.schema, self.table)

    @classmethod
    def parse_int(cls, row, pos, size):
        if size - pos < 8:
            raise InvalidRowData('insufficient bytes to decode value')

        u = read_be_uint64(row[pos:pos + 8])
        v = SIGNMASK ^ u
        return v, 8

    @classmethod
    def parse_uint(cls, row, pos, size):
        if size - pos < 8:
            raise InvalidRowData('insufficient bytes to decode value')

        v = read_be_uint64(row[pos:pos + 8])
        return v, 8

    @classmethod
    def parse_varint(cls, row, pos, size):
        v, n = read_varint(row[pos:pos + 10])  # needs at most 10 bytes
        if n > 0:
            return v, n

        if n < 0:
            raise InvalidRowData("value larger than 64 bits")

        raise InvalidRowData("insufficient bytes to decode value")

    @classmethod
    def parse_uvarint(cls, row, pos, size):
        v, n = read_uvarint(row[pos:pos + 10])  # needs at most 10 bytes
        if n > 0:
            return v, n

        if n < 0:
            raise InvalidRowData("value larger than 64 bits")

        raise InvalidRowData("insufficient bytes to decode value")

    @classmethod
    def parse_float(cls, row, pos, size):
        if size - pos < 8:
            raise InvalidRowData('insufficient bytes to decode value')

        tmp = bytearray(row[pos:pos + 8])
        if tmp[0] & 0x80 > 0:
            tmp[0] &= 0x7F
        else:
            u = struct.unpack(b">Q", str(tmp))[0]
            u = ~u
            tmp = struct.pack(b">q", u)

        v = struct.unpack(b">d", str(tmp))[0]
        return v, 8

    @classmethod
    def parse_bytes(cls, row, pos, size):
        ENC_GROUP_SIZE = 8
        ENC_MARKER = 0xFF
        ENC_PAD = 0x0

        old_pos = pos

        retval = ""
        while True:
            if size - pos < ENC_GROUP_SIZE + 1:
                raise InvalidRowData("insufficient bytes to decode value")

            group = row[pos:pos + ENC_GROUP_SIZE]
            marker = row[pos + ENC_GROUP_SIZE]

            pad_count = ENC_MARKER - marker

            if pad_count > ENC_GROUP_SIZE:
                raise InvalidRowData("invalid marker byte")

            real_roup_size = ENC_GROUP_SIZE - pad_count
            retval += group[:real_roup_size]
            pos += ENC_GROUP_SIZE + 1

            if pad_count != 0:
                pad_byte = ENC_PAD

                for v in group[real_roup_size:]:
                    if v != pad_byte:
                        raise InvalidRowData("invalid padding byte")

                break

        return retval, pos - old_pos

    @classmethod
    def parse_compact_bytes(cls, row, pos, size):
        n, s = cls.parse_varint(row, pos, size)
        if size - pos - s < n:
            raise InvalidRowData(
                "insufficient bytes to decode value, expected length: %s", n)

        return row[pos + s:pos + s + n], s + n

    @classmethod
    def parse_decimal(cls, row, pos, size):
        if size - pos < 3:
            raise InvalidRowData('insufficient bytes to decode value')

        precision = read_int8(row[pos])
        frac = read_int8(row[pos + 1])

        bin_size = decimal_bin_size(precision, frac)

        if size - pos < bin_size + 2:
            raise InvalidRowData("insufficient bytes to decode value")

        bin = row[pos + 2:pos + 2 + bin_size]
        bin_pos = 0

        if read_int8(bin[bin_pos]) & 0x80 > 0:
            negitive = False
            retval = ''
        else:
            negitive = True
            retval = '-'

        bin = bytearray(bin)
        bin[0] ^= 0x80
        bin = str(bin)

        # The number of *decimal* digits before the point.
        digits_int = precision - frac
        # The number of 32bit words before the point.
        words_int = digits_int / DIGITS_PER_WORD
        # The number of leading *decimal* digits not in a word
        leading_digits = digits_int - words_int * DIGITS_PER_WORD

        # The number of 32bit words after the point.
        words_frac = frac / DIGITS_PER_WORD
        # The number of trailing *decimal* digits not in a word
        trailing_digits = frac - words_frac * DIGITS_PER_WORD

        if leading_digits > 0:
            i = DIG2BYTES[leading_digits]
            x = read_int32_word(bin[bin_pos:], i)
            x = ~x if negitive else x
            retval += str(x)
            bin_pos += i

        for i in range(0, words_int * WORD_SIZE, WORD_SIZE):
            x = read_int32_word(bin[bin_pos + i:], WORD_SIZE)
            x = ~x if negitive else x
            retval += str(x)

        bin_pos += words_int * WORD_SIZE

        retval += '.'

        for i in range(0, words_frac * WORD_SIZE, WORD_SIZE):
            x = read_int32_word(bin[bin_pos + i:], WORD_SIZE)
            x = ~x if negitive else x
            retval += str(x)

        bin_pos += words_frac * WORD_SIZE

        if trailing_digits > 0:
            i = DIG2BYTES[trailing_digits]
            x = read_int32_word(bin[bin_pos:], i)
            x = ~x if negitive else x
            retval += str(x)

        return retval, bin_size + 2

    @classmethod
    def parse_duration(cls, row, pos, size):
        v, s = cls.parse_int(row, pos, size)
        d = timedelta(microseconds=v / 1000)

        return d, s

    @classmethod
    def parse_one_column(cls, value):
        pos = 0
        size = len(value)

        if size - pos < 1:
            raise InvalidRowData("Invalid encoded key")

        flag = read_int8(value[pos])
        pos += 1

        if cls.INT_FLAG == flag:
            v, _ = cls.parse_int(value, pos, size)
        elif cls.UINT_FLAG == flag:
            v, _ = cls.parse_uint(value, pos, size)
        elif cls.VARINT_FLAG == flag:
            v, _ = cls.parse_varint(value, pos, size)
        elif cls.UVARINT_FLAG == flag:
            v, _ = cls.parse_uvarint(value, pos, size)
        elif cls.FLOAT_FLAG == flag:
            v, _ = cls.parse_float(value, pos, size)
        elif cls.BYTES_FLAG == flag:
            v, _ = cls.parse_bytes(value, pos, size)
        elif cls.COMPACTBYTES_FLAG == flag:
            v, _ = cls.parse_compact_bytes(value, pos, size)
        elif cls.DECIMAL_FLAG == flag:
            v, _ = cls.parse_decimal(value, pos, size)
        elif cls.DURATION_FLAG == flag:
            v, _ = cls.parse_duration(value, pos, size)
        elif cls.NIL_FLAG == flag:
            v = None
        else:
            raise InvalidRowData("Invalid encoded key")

        return v

    @classmethod
    def parse_insert_and_delete_row(cls, row):
        values = {}
        types = {}

        for c in row:
            col = column.FromString(c)
            tp = read_uint8(col.tp)
            mysql_type = col.mysql_type
            name = get_unicode(col.name)

            value = cls.parse_one_column(col.value)

            if value is not None and \
                    mysql_type in (
                        MYSQL_TYPE_TINYTEXT,
                        MYSQL_TYPE_MEDIUMTEXT,
                        MYSQL_TYPE_LONGTEXT,
                        MYSQL_TYPE_TEXT,
                        MYSQL_TYPE_CHAR,
                        MYSQL_TYPE_VARCHAR
                    ):
                value = value.decode('utf-8')

            values[name] = value
            types[name] = (tp, mysql_type)

        return types, values

    @classmethod
    def parse_update_row(cls, row):
        old_values = {}
        new_values = {}
        types = {}

        for c in row:
            col = column.FromString(c)
            tp = read_uint8(col.tp)
            mysql_type = col.mysql_type
            name = unicode(col.name)

            value = cls.parse_one_column(col.value)
            changed_value = cls.parse_one_column(col.changed_value)

            old_values[name] = value
            new_values[name] = changed_value
            types[name] = (tp, mysql_type)

        return types, old_values, new_values


class DeleteRowsEvent(RowsEvent):
    """This event is trigger when a row in the database is removed

    For each row you have a hash with a single key:
    values which contain the data of the removed line.
    """

    def __init__(self, binlog, event):
        super(DeleteRowsEvent, self).__init__(binlog, event)
        self._types, self._values = self.parse_insert_and_delete_row(event.row)

    @property
    def types(self):
        return self._types

    @property
    def values(self):
        return self._values

    def __str__(self):
        return unicode(self).encode('utf-8')

    def __unicode__(self):
        parent = super(DeleteRowsEvent, self).__unicode__()
        values = self.values
        types = self.types

        s = ''
        for col in values:
            s += "%s %s, " % (col, format_column(types[col], values[col]))

        return "%s: %s" % (parent, s)


class WriteRowsEvent(RowsEvent):
    """This event is triggered when a row in database is added

    For each row you have a hash with a single key:
    values which contain the data of the new line.
    """

    def __init__(self, binlog, event):
        super(WriteRowsEvent, self).__init__(binlog, event)
        self._types, self._values = self.parse_insert_and_delete_row(event.row)

    @property
    def types(self):
        return self._types

    @property
    def values(self):
        return self._values

    def __str__(self):
        return unicode(self).encode('utf-8')

    def __unicode__(self):
        parent = super(WriteRowsEvent, self).__unicode__()
        values = self.values
        types = self.types

        s = ''
        for col in values:
            s += "%s %s, " % (col, format_column(types[col], values[col]))

        return "%s: %s" % (parent, s)


class UpdateRowsEvent(RowsEvent):
    """This event is triggered when a row in the database is changed

    For each row you got a hash with two keys:
        * before_values
        * after_values
    """

    def __init__(self, binlog, event):
        super(UpdateRowsEvent, self).__init__(binlog, event)
        self._types, self._before_values, self._after_values = \
            self.parse_update_row(event.row)

    @property
    def types(self):
        return self._types

    @property
    def before_values(self):
        return self._before_values

    @property
    def after_values(self):
        return self._after_values

    def __str__(self):
        return unicode(self).encode('utf-8')

    def __unicode__(self):
        parent = super(UpdateRowsEvent, self).__unicode__()
        before_values = self.before_values
        after_values = self.after_values
        types = self.types

        s = ''
        for col in before_values:
            s += "%s %s => %s, " % (
                col, format_column(types[col], before_values[col]),
                format_column(types[col], after_values[col]))

        return "%s: %s" % (parent, s)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pytidbrep/pb_binlog/binlog.proto

import sys
_b = sys.version_info[0] < 3 and (lambda x: x) or (
    lambda x: x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()

DESCRIPTOR = _descriptor.FileDescriptor(
    name='pytidbrep/pb_binlog/binlog.proto',
    package='pb_binlog',
    syntax='proto2',
    serialized_pb=
    _b('\n pytidbrep/pb_binlog/binlog.proto\x12\tpb_binlog\"\\\n\x06\x63olumn\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\n\n\x02tp\x18\x02 \x01(\x0c\x12\x12\n\nmysql_type\x18\x03 \x01(\t\x12\r\n\x05value\x18\x04 \x01(\x0c\x12\x15\n\rchanged_value\x18\x05 \x01(\x0c\"_\n\x05\x45vent\x12\x13\n\x0bschema_name\x18\x01 \x01(\t\x12\x12\n\ntable_name\x18\x02 \x01(\t\x12 \n\x02tp\x18\x03 \x01(\x0e\x32\x14.pb_binlog.EventType\x12\x0b\n\x03row\x18\x04 \x03(\x0c\"+\n\x07\x44MLData\x12 \n\x06\x65vents\x18\x01 \x03(\x0b\x32\x10.pb_binlog.Event\"w\n\x06\x42inlog\x12!\n\x02tp\x18\x01 \x01(\x0e\x32\x15.pb_binlog.BinlogType\x12\x11\n\tcommit_ts\x18\x02 \x01(\x03\x12$\n\x08\x64ml_data\x18\x03 \x01(\x0b\x32\x12.pb_binlog.DMLData\x12\x11\n\tddl_query\x18\x04 \x01(\x0c*/\n\tEventType\x12\n\n\x06Insert\x10\x00\x12\n\n\x06Update\x10\x01\x12\n\n\x06\x44\x65lete\x10\x02*\x1e\n\nBinlogType\x12\x07\n\x03\x44ML\x10\x00\x12\x07\n\x03\x44\x44L\x10\x01'
       ))
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

_EVENTTYPE = _descriptor.EnumDescriptor(
    name='EventType',
    full_name='pb_binlog.EventType',
    filename=None,
    file=DESCRIPTOR,
    values=[
        _descriptor.EnumValueDescriptor(
            name='Insert', index=0, number=0, options=None, type=None),
        _descriptor.EnumValueDescriptor(
            name='Update', index=1, number=1, options=None, type=None),
        _descriptor.EnumValueDescriptor(
            name='Delete', index=2, number=2, options=None, type=None),
    ],
    containing_type=None,
    options=None,
    serialized_start=404,
    serialized_end=451, )
_sym_db.RegisterEnumDescriptor(_EVENTTYPE)

EventType = enum_type_wrapper.EnumTypeWrapper(_EVENTTYPE)
_BINLOGTYPE = _descriptor.EnumDescriptor(
    name='BinlogType',
    full_name='pb_binlog.BinlogType',
    filename=None,
    file=DESCRIPTOR,
    values=[
        _descriptor.EnumValueDescriptor(
            name='DML', index=0, number=0, options=None, type=None),
        _descriptor.EnumValueDescriptor(
            name='DDL', index=1, number=1, options=None, type=None),
    ],
    containing_type=None,
    options=None,
    serialized_start=453,
    serialized_end=483, )
_sym_db.RegisterEnumDescriptor(_BINLOGTYPE)

BinlogType = enum_type_wrapper.EnumTypeWrapper(_BINLOGTYPE)
Insert = 0
Update = 1
Delete = 2
DML = 0
DDL = 1

_COLUMN = _descriptor.Descriptor(
    name='column',
    full_name='pb_binlog.column',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='name',
            full_name='pb_binlog.column.name',
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='tp',
            full_name='pb_binlog.column.tp',
            index=1,
            number=2,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b(""),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='mysql_type',
            full_name='pb_binlog.column.mysql_type',
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='value',
            full_name='pb_binlog.column.value',
            index=3,
            number=4,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b(""),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='changed_value',
            full_name='pb_binlog.column.changed_value',
            index=4,
            number=5,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b(""),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[],
    serialized_start=47,
    serialized_end=139, )

_EVENT = _descriptor.Descriptor(
    name='Event',
    full_name='pb_binlog.Event',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='schema_name',
            full_name='pb_binlog.Event.schema_name',
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='table_name',
            full_name='pb_binlog.Event.table_name',
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='tp',
            full_name='pb_binlog.Event.tp',
            index=2,
            number=3,
            type=14,
            cpp_type=8,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='row',
            full_name='pb_binlog.Event.row',
            index=3,
            number=4,
            type=12,
            cpp_type=9,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[],
    serialized_start=141,
    serialized_end=236, )

_DMLDATA = _descriptor.Descriptor(
    name='DMLData',
    full_name='pb_binlog.DMLData',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='events',
            full_name='pb_binlog.DMLData.events',
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[],
    serialized_start=238,
    serialized_end=281, )

_BINLOG = _descriptor.Descriptor(
    name='Binlog',
    full_name='pb_binlog.Binlog',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='tp',
            full_name='pb_binlog.Binlog.tp',
            index=0,
            number=1,
            type=14,
            cpp_type=8,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='commit_ts',
            full_name='pb_binlog.Binlog.commit_ts',
            index=1,
            number=2,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='dml_data',
            full_name='pb_binlog.Binlog.dml_data',
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='ddl_query',
            full_name='pb_binlog.Binlog.ddl_query',
            index=3,
            number=4,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b(""),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[],
    serialized_start=283,
    serialized_end=402, )

_EVENT.fields_by_name['tp'].enum_type = _EVENTTYPE
_DMLDATA.fields_by_name['events'].message_type = _EVENT
_BINLOG.fields_by_name['tp'].enum_type = _BINLOGTYPE
_BINLOG.fields_by_name['dml_data'].message_type = _DMLDATA
DESCRIPTOR.message_types_by_name['column'] = _COLUMN
DESCRIPTOR.message_types_by_name['Event'] = _EVENT
DESCRIPTOR.message_types_by_name['DMLData'] = _DMLDATA
DESCRIPTOR.message_types_by_name['Binlog'] = _BINLOG
DESCRIPTOR.enum_types_by_name['EventType'] = _EVENTTYPE
DESCRIPTOR.enum_types_by_name['BinlogType'] = _BINLOGTYPE

column = _reflection.GeneratedProtocolMessageType(
    'column',
    (_message.Message, ),
    dict(
        DESCRIPTOR=_COLUMN,
        __module__='pytidbrep.pb_binlog.binlog_pb2'
        # @@protoc_insertion_point(class_scope:pb_binlog.column)
    ))
_sym_db.RegisterMessage(column)

Event = _reflection.GeneratedProtocolMessageType(
    'Event',
    (_message.Message, ),
    dict(
        DESCRIPTOR=_EVENT,
        __module__='pytidbrep.pb_binlog.binlog_pb2'
        # @@protoc_insertion_point(class_scope:pb_binlog.Event)
    ))
_sym_db.RegisterMessage(Event)

DMLData = _reflection.GeneratedProtocolMessageType(
    'DMLData',
    (_message.Message, ),
    dict(
        DESCRIPTOR=_DMLDATA,
        __module__='pytidbrep.pb_binlog.binlog_pb2'
        # @@protoc_insertion_point(class_scope:pb_binlog.DMLData)
    ))
_sym_db.RegisterMessage(DMLData)

Binlog = _reflection.GeneratedProtocolMessageType(
    'Binlog',
    (_message.Message, ),
    dict(
        DESCRIPTOR=_BINLOG,
        __module__='pytidbrep.pb_binlog.binlog_pb2'
        # @@protoc_insertion_point(class_scope:pb_binlog.Binlog)
    ))
_sym_db.RegisterMessage(Binlog)

# @@protoc_insertion_point(module_scope)



# -*- coding:utf-8 -*-

import unittest

from pytidbrep.event import RowsEvent
from pytidbrep.event import read_int32_word


class TestEvent(unittest.TestCase):

    def test_read_int32_word(self):
        row = '112233'.decode("hex")
        v = read_int32_word(row, 3)
        assert 1122867 == v

        row = 'EEDDCC'.decode("hex")
        v = read_int32_word(row, 3)
        v = ~v
        assert 1122867 == v

    def test_parse_int(self):
        pass

    def test_parse_uint(self):
        pass

    def test_read_varint(self):
        pass

    def test_read_uvarint(self):
        pass

    def test_float(self):
        pass

    def test_parse_decimal(self):
        row = '0DFB38D204D2'.decode("hex")
        v = read_int32_word(row, 4)
        assert 234567890 == v

        row = '0E04810DFB38D204D2'.decode("hex")
        v, s = RowsEvent.parse_decimal(row, 0, len(row))
        assert '1234567890.1234' == v
        assert len(row) == s

        row = '0E047EF204C72DFB2D'.decode("hex")
        v, s = RowsEvent.parse_decimal(row, 0, len(row))
        assert '-1234567890.1234' == v
        assert len(row) == s


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



# -*- coding: utf-8 -*-

# Copyright (c) 2016 HashData Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from __future__ import unicode_literals

import logging
import os
import re
import struct
import time

from pytidbrep.crc32c import crc32c
from pytidbrep.event import DDLEvent
from pytidbrep.event import DeleteRowsEvent
from pytidbrep.event import UpdateRowsEvent
from pytidbrep.event import WriteRowsEvent
from pytidbrep.event import XidEvent
from pytidbrep.expcetion import CrcNotMatcheError
from pytidbrep.expcetion import EofError
from pytidbrep.expcetion import InvalidRowData
from pytidbrep.expcetion import MagicNotMatchError
from pytidbrep.expcetion import UnexpectedEofError
from pytidbrep.expcetion import UnknownBinlogType
from pytidbrep.expcetion import UnknownDMLType
from pytidbrep.pb_binlog.binlog_pb2 import Binlog
from pytidbrep.pb_binlog.binlog_pb2 import DDL
from pytidbrep.pb_binlog.binlog_pb2 import DML
from pytidbrep.pb_binlog.binlog_pb2 import Delete
from pytidbrep.pb_binlog.binlog_pb2 import Insert
from pytidbrep.pb_binlog.binlog_pb2 import Update

LOG = logging.getLogger(__name__)


class BinLogStreamReader(object):
    # the file name format is like binlog-0000000000000001
    binlog_name_patten = r"binlog-[0-9]{16}"
    binlog_name_matcher = re.compile(binlog_name_patten)
    binlog_magic = 471532804

    def __init__(self,
                 binlog_dir,
                 binlog_start=None,
                 blocking=False,
                 skip_to_timestamp=None,
                 ignore_error=False):
        self._binlog_dir = binlog_dir
        self._current_binlog_file = None
        self._current_binlog_file_name = None
        self._current_binlog_file_position = 0
        self._skip_to_timestamp = long(skip_to_timestamp)
        self._blocking = blocking
        self._ignore_error = ignore_error
        self._current_events = []

        if binlog_start:
            self._current_binlog_file = self._get_next_binlog_file(
                binlog_start)

    @property
    def current_binlog_file_name(self):
        return self._current_binlog_file_name

    def _read_fully(self, size):
        b = b''
        s = size

        while s > 0:
            if self._current_binlog_file:
                d = self._current_binlog_file.read(s)
            else:
                d = None

            if not d:
                next_file = self._get_next_binlog_file()

                if next_file:
                    if s == size:
                        LOG.info('Reading binlog "%s"' %
                                 self._current_binlog_file_name)
                        if self._current_binlog_file:
                            self._current_binlog_file.close()
                        self._current_binlog_file = next_file
                    else:
                        # binlog event will not cross file boundary
                        raise UnexpectedEofError(
                            'Unexpected end of binlog file "%s"' %
                            self._current_binlog_file_name)
                else:
                    if self._blocking:
                        self._current_binlog_file.seek(
                            self._current_binlog_file_position)
                        time.sleep(1)
                    elif s == size:
                        raise EofError('End of binlog file "%s"' %
                                       self._current_binlog_file_name)
                    else:
                        raise UnexpectedEofError(
                            'Unexpected end of binlog file "%s"' %
                            self._current_binlog_file_name)

                continue

            l = len(d)
            s -= l
            b += d
            self._current_binlog_file_position += l

        return b

    def _read_le_int32(self):
        b = self._read_fully(4)
        return struct.unpack(b"<i", b)[0]

    def _read_le_uint32(self):
        b = self._read_fully(4)
        return struct.unpack(b"<I", b)[0]

    def _read_le_int64(self):
        b = self._read_fully(8)
        return struct.unpack(b"<q", b)[0]

    def _get_next_binlog_file(self, start_file=None):
        if not os.path.exists(self._binlog_dir):
            raise RuntimeError(
                'binlog directory "%s" does not exist' % self._binlog_dir)

        binlogs = [
            f for f in os.listdir(self._binlog_dir)
            if os.path.isfile(os.path.join(self._binlog_dir, f)) and
            self.binlog_name_matcher.match(f)
        ]

        binlogs.sort()

        # with specified start position in directory
        if start_file:
            binlogs.index(start_file)
            self._current_binlog_file_name = start_file
            self._current_binlog_file_position = 0
            return open(
                os.path.join(self._binlog_dir, self._current_binlog_file_name),
                'rb')

        if not binlogs:
            return None

        # start with the first file in the directory
        if not self._current_binlog_file_name:
            self._current_binlog_file_name = binlogs[0]
            self._current_binlog_file_position = 0
            return open(
                os.path.join(self._binlog_dir, self._current_binlog_file_name),
                'rb')

        # next file in the directory
        index = binlogs.index(self._current_binlog_file_name)

        if index + 1 < len(binlogs):
            self._current_binlog_file_name = binlogs[index + 1]
            self._current_binlog_file_position = 0
            return open(
                os.path.join(self._binlog_dir, self._current_binlog_file_name),
                'rb')
        else:
            return None

    def _read_payload(self):
        try:
            magic = self._read_le_int32()
        except EofError:
            return None

        if magic != self.binlog_magic:
            raise MagicNotMatchError(
                '%s is not a valid binlog file: Magic not match')

        size = self._read_le_int64()
        payload = self._read_fully(size)
        crc = self._read_le_uint32()

        if crc != crc32c(payload):
            raise CrcNotMatcheError('CRC32 does not match in file "%s"' %
                                    self._current_binlog_file_name)

        return payload

    def _decode_DML(self, binlog):
        data = binlog.dml_data

        for event in data.events:
            if event.tp == Insert:
                self._current_events.append(WriteRowsEvent(binlog, event))
            elif event.tp == Update:
                self._current_events.append(UpdateRowsEvent(binlog, event))
            elif event.tp == Delete:
                self._current_events.append(DeleteRowsEvent(binlog, event))
            else:
                raise UnknownDMLType("Unknown DML type: %s" % event.tp)

        self._current_events.append(XidEvent(binlog.commit_ts))

    def fetchone(self):
        while True:
            if self._current_events:
                return self._current_events.pop(0)

            payload = self._read_payload()

            if not payload:
                return None

            try:
                binlog = Binlog.FromString(payload)

                if self._skip_to_timestamp and \
                   binlog.commit_ts < self._skip_to_timestamp:
                    continue

                if binlog.tp == DML:
                    self._decode_DML(binlog)
                    continue
                elif binlog.tp == DDL:
                    self._current_events.append(DDLEvent(binlog))
                    self._current_events.append(XidEvent(binlog.commit_ts))
                    continue
                else:
                    raise UnknownBinlogType(
                        'unknown binlog type %s' % binlog.tp)
            except InvalidRowData:
                if self._ignore_error:
                    LOG.warning('Ignore invalid binlog entry at ts: %s' %
                                binlog.commit_ts)
                else:
                    raise

    def __iter__(self):
        return iter(self.fetchone, None)

    def close(self):
        self._current_binlog_file.close()


if __name__ == "__main__":
    logging.basicConfig()
    stream = BinLogStreamReader(
        '/hashdata/pytidbrep/data/data/drainer',
        skip_to_timestamp=0,
        ignore_error=True,
        blocking=False)
    for event in stream:
        print event
